package edu.berkeley.qfat.bind;
import edu.berkeley.qfat.geom.*;
import java.util.*;

/**
 *  An equivalence class of geometric objects whose positions are
 *  related by affine transformation matrices and are constrained by
 *  an affine constraint.
 */
class BindingGroup<T extends HasBindingGroup> implements Iterable<T> {

    /** the arbitrarily-chosen master of the binding group */
    private T                  master     = null;

    /** the affine constraint, in master coordinates, of this binding group */
    private AffineConstraint   constraint = AffineConstraint.ALL;

    /**
     *  For each member of the binding group, the matrix which must be
     *  multiplied by the master to get the member's position
     */
    private HashMap<T,Matrix>  matrices   = new HashMap<T,Matrix>();

    BindingGroup(T master) {
        this.master = master;
        matrices.put(master, Matrix.ONE);
    }

    /** the size of this binding group */
    int size() { return matrices.size(); }

    /** merge another binding group into this one */
    void merge(BindingGroup<T> bg, Matrix m, float epsilon) {
        if (bg==this) {
            if (m.equalsModuloEpsilon(Matrix.ONE, epsilon)) return;
            /*
            if (master instanceof edu.berkeley.qfat.Mesh.Vertex)
                System.err.println(m.times(m));
            */
            // FIXME: what if points do not fall on the merged constraint-line?
            constraint = constraint.intersect(m.getAffineConstraint(epsilon), epsilon);
            return;
        }

        // bg.master = m * master
        // hbg       = bg.getMatrix(hbg) * bg.master
        // hbg       = bg.getMatrix(hbg) * m * master
        for(HasBindingGroup hbg : bg) {
            matrices.put((T)hbg, bg.getMatrix((T)hbg).times(m));
            hbg.bindingGroup = this;
        }

        // FIXME: what if points do not fall on the merged constraint-line?
        AffineConstraint ac = bg.constraint.multiply(getMatrix(master, bg.master));
        constraint = constraint.intersect(ac, epsilon);

        bg.master = null;
        for(HasBindingGroup hbg : bg) hbg.bindingGroupChanged();
        bg.matrices.clear();
        bg.matrices = null;
    }

    Matrix getMatrix(T t) { return matrices.get(t); }

    public Iterator<T> iterator() { return matrices.keySet().iterator(); }

    /** t1 = getMatrix(t1, t2) * t2 */
    Matrix getMatrix(T t1, T t2) {
        //                    t1 = getMatrix(t1) * master
        // getMatrix(t2)^-1 * t2 =                 master
        //                    t1 = getMatrix(t1) * getMatrix(t2)^-1 * t2
        return getMatrix(t1).times(getMatrix(t2).inverse());
    }

    AffineConstraint getAffineConstraint(T t) {
        return constraint.multiply(matrices.get(t));
    }

    void unbind(T trem) {
        if (trem != master) {
            matrices.remove(trem);
            return;
        }
        if (matrices.size()==1) {
            master = null;
            matrices.remove(trem);
            return;
        }
        Iterator<T> it = iterator();
        T newmaster = it.next();
        if (newmaster==trem) newmaster = it.next();
        if (newmaster==trem) throw new Error("impossible");

        // FIXME: is this correct?
        constraint = constraint.multiply(getMatrix(newmaster, master));

        HashMap<T,Matrix> newmatrices = new HashMap<T,Matrix>();
        for(T t : matrices.keySet()) {
            if (t==trem) continue;
            newmatrices.put(t, getMatrix(t, newmaster));
        }
        master = newmaster;
        matrices = newmatrices;
    }

    boolean contains(HasBindingGroup t) {
        return matrices.get((T)t) != null;
    }
}
